from datetime import datetime
import pytz

DATE_PREFIX = '/Date('
DATE_SUFFIX = ')/'

ELVIS_TIMEZONE = pytz.timezone('Europe/Tallinn')
UTC = pytz.timezone('UTC')


def decode_elvis_timestamp(timestamp: str):
    """Try to convert the argument to timestamp using ELVIS rules, return it unmodified if impossible"""
    str_timestamp = str(timestamp).strip()
    if str_timestamp.startswith(DATE_PREFIX) and str_timestamp.endswith(DATE_SUFFIX):
        milliseconds = str_timestamp[len(DATE_PREFIX):-len(DATE_SUFFIX)]
        timezone_offset = 0
        try:
            if "+" in milliseconds:
                timezone_offset_string = milliseconds[milliseconds.index("+")+1:]
                milliseconds = milliseconds[:milliseconds.index("+")]
                if len(timezone_offset_string) == 4:
                    timezone_offset = int(timezone_offset_string[:2])*60+int(timezone_offset_string[2:])
            seconds = int(milliseconds) / 1000
        except ValueError:
            return timestamp

        # Elvis Timezone offsets are relevant to Elvis natural timezone (Tallinn)
        return ELVIS_TIMEZONE.localize(datetime.fromtimestamp(seconds).astimezone(
            pytz.FixedOffset(-timezone_offset)
        ).replace(tzinfo=None))

    return timestamp



# coding:utf-8

import base64
from copy import copy
import json
import requests
from datetime import datetime
from decimal import Decimal

from django.utils.encoding import force_text

from .enums import WarehouseType, AssortmentType, TransportOrderRoleContext, WaybillRoleContext
from .models import (
    FilterItem, SortItem, Address, ElvisModel, TransportOrderListPage, TransportOrder, TransportOrderStatusInfo,
    TimberWarehouse, Waybill, WaybillStatusInfo, WaybillListPage, TimberAssortment, FineMeasurementFile,
)
from .utils import DATE_PREFIX, DATE_SUFFIX, ELVIS_TIMEZONE


class ElvisException(Exception):
    def __init__(self, message, raw):
        self.message = message
        self.raw = raw

    @classmethod
    def reformat_elvis_exception_text(cls, value, indent=1):
        new_value = value

        if isinstance(value, dict):
            new_value = "\n"
            for key in value:
                new_value = "%s%s`%s`:%s\n" % (
                    new_value,
                    "    " * indent,
                    key,
                    cls.reformat_elvis_exception_text(
                        value[key], indent=indent+1
                    ),
                )

        elif isinstance(value, list):
            new_value = "\n"
            for index, item in enumerate(value):
                new_value = "%s%s[%s]:%s\n" % (
                    new_value,
                    "    " * indent,
                    index,
                    cls.reformat_elvis_exception_text(
                        item, indent=indent+1
                    ),
                )

        elif hasattr(value, "replace"):
            new_value = value.replace(u"\\u000d\\u000a", "\n")

        return new_value

    def __str__(self):
        return "ElvisException %s: %s" % (self.message, self.reformat_elvis_exception_text(self.raw))


class ElvisEncoder(json.JSONEncoder):
    ELVIS_OBJECTS = (
        FilterItem,
        SortItem,
        WarehouseType,
        Address,
        ElvisModel
    )

    def default(self, obj):
        if isinstance(obj, Decimal):
            return float(obj)

        if isinstance(obj, datetime):
            # Elvis timestamp offsets are from Tallinn time
            timezone_offset = obj.utcoffset() - obj.astimezone(ELVIS_TIMEZONE).utcoffset()

            timezone_offset_hours = timezone_offset.total_seconds() // 3600
            timezone_offset_minutes = (timezone_offset.total_seconds() - timezone_offset_hours * 3600) // 60
            return '%s%d+%02d%02d%s' % (
                DATE_PREFIX,
                (obj - datetime(1970, 1, 1)).total_seconds() * 1000,
                timezone_offset_hours,
                timezone_offset_minutes,
                DATE_SUFFIX,
            )

        if not isinstance(obj, ElvisEncoder.ELVIS_OBJECTS):
            return super(ElvisEncoder, self).default(obj)

        ret = copy(obj.__dict__)

        if isinstance(obj, FineMeasurementFile):
            del ret['Data']

        if 'Id' in ret and ret['Id'] is None:
            del ret['Id']

        if '_already_loaded' in ret:
            del ret['_already_loaded']

        return ret


class ElvisClient(object):
    DEFAULT_REQUEST_TIMEOUT = 120  # 120 seconds

    def __init__(
        self, api_url, person_code, certificate_pass="",
        session_token=None, request_timeout=DEFAULT_REQUEST_TIMEOUT,
    ):
        self.api_url = api_url

        self.person_code = person_code
        self.certificate = None
        self.certificate_password = certificate_pass or ""
        self.session_token = session_token or None
        self.request_timeout = request_timeout

        assert self.person_code, "Person code must be provided"

    def load_cert_from_fieldfile(self, fieldfile):
        try:
            fieldfile.open(mode='rb')
            self.certificate = force_text(base64.b64encode(fieldfile.read()), 'utf-8')
            fieldfile.close()
        except IOError:
            raise Exception("FieldFile %s is invalid" % fieldfile)

    def load_cert_from_file(self, file_path):
        try:
            with open(file_path, "rb") as fHandle:
                self.certificate = force_text(base64.b64encode(fHandle.read()), 'utf-8')
                fHandle.close()
        except IOError:
            raise Exception("File %s is invalid" % file_path)

    def load_cert_from_str(self, certificate_data):
        # Note: may need validation here?
        self.certificate = force_text(base64.b64encode(certificate_data), 'utf-8')

    def load_cert_from_base64str(self, certificate_data):
        # Note: may need validation here?
        self.certificate = certificate_data

    def __request(self, endpoint, method, attrs=None):

        func = requests.get
        if method.lower() == "post":
            func = requests.post

        headers = {
            'Content-type': 'application/json',
            'Accept': 'text/plain',
        }

        if self.session_token:
            headers['Authorization'] = self.session_token

        if method.lower() == "post":
            result = func(self.api_url % endpoint, data=json.dumps(attrs, cls=ElvisEncoder), headers=headers,
                          timeout=self.request_timeout)
        else:
            result = func(self.api_url % endpoint, params=attrs, headers=headers, timeout=self.request_timeout)

        error_message = ""
        try:
            json_data = json.loads(result.text)
        except ValueError:
            error_message = "not json"
            json_data = result.text

        if result.status_code == 200:
            success = True
        else:
            error_message = "Bad status code: %d" % result.status_code,
            success = False

        result = {
            "Success": success,
            "raw": json_data,
        }

        if error_message:
            result["message"] = error_message

        return result

    # ENDPOINTS

    def server_info(self):
        assert self.session_token, "No valid session available"

        result = self.__request("GetInformation", "GET")

        if result["Success"] and result["raw"].get("GetInformationResult", None):
            return result["raw"]["GetInformationResult"]
        else:
            raise ElvisException("Server information not available, it was added in proxy version 1.1.*", raw=result)

    def authorize(self, session_tag='default'):
        assert self.certificate, "Certificate must be provided"

        result = self.__request("Authorize", "POST", {
            'code': self.person_code,
            'certificateData': self.certificate,
            'password': self.certificate_password,
            'connectionTag': session_tag,
        })

        if result["Success"] and result["raw"].get("AuthorizeResult", None) and result['raw']['AuthorizeResult'].get('Success'):
            auth_result = result["raw"].get("AuthorizeResult")
            self.session_token = auth_result["access_token"]
            return True
        else:
            raise ElvisException("Authorization failed:", result)

    def get_session_tag(self):
        assert self.session_token, "No valid session available"

        result = self.__request("getSessionTag", "GET", {})

        if result["Success"]:
            return result["raw"]['getSessionTagResult']
        else:
            raise ElvisException(result['message'], result['raw'])

    def get_assortment_types(self, assortment_type=AssortmentType.ELVIS):
        assert self.session_token, "No valid session available"
        assert assortment_type in [AssortmentType.ELVIS, AssortmentType.COMPANY], "Invalid assortment type"

        result = self.__request("getAssortmentTypes", "GET", {
            'assortment_type': assortment_type,
        })

        if result["Success"]:
            return result["raw"]['getAssortmentTypesResult']
        else:
            raise ElvisException(result['message'], result['raw'])

    def search_warehouses(self, filters, sorting, start=0, limit=10, show_count=False):
        assert self.session_token, "No valid session available"

        assert isinstance(filters, (list, tuple, FilterItem)), 'Filters must be a list or tuple'
        if isinstance(filters, FilterItem):
            filters = (filters, )

        assert not list(filter(lambda x: not isinstance(x, FilterItem), filters)), \
            'All filters must be instances of FilterItem'

        assert isinstance(sorting, (list, tuple, SortItem)), 'Sorting must be a list or tuple'
        if isinstance(sorting, SortItem):
            sorting = (sorting, )

        assert not list(filter(lambda x: not isinstance(x, SortItem), sorting)), \
            'All sorting rules must be instances of SortItem'

        assert filters and sorting, 'Need to add filters and sorting!'

        result = self.__request("SearchWarehouses", "POST", {
            'filters': filters,
            'sorting': sorting,
            'start': start,
            'limit': limit,
            'show_count': show_count,
        })

        if result["Success"]:
            return result["raw"]["SearchWarehousesResult"]
        else:
            raise ElvisException(result['message'], result['raw'])

    def insert_transport_order(self, transport_order):
        assert self.session_token, "No valid session available"

        assert isinstance(transport_order, TransportOrder), 'Invalid TransportOrder'

        result = self.__request("InsertTransportOrder", "POST", {
            'item': transport_order,
        })

        if result.get("Success", False):
            return result["raw"]['InsertTransportOrderResult']
        else:
            raise ElvisException(result['message'], result['raw'])

    def search_transport_orders(self, context, filters, sorting, start=0, limit=10, show_count=False):
        assert self.session_token, "No valid session available"

        assert context in TransportOrderRoleContext.__dict__.values(), 'Invalid context'

        assert isinstance(filters, (list, tuple, FilterItem)), 'Filters must be a list or tuple'
        if isinstance(filters, FilterItem):
            filters = (filters, )

        assert not list(filter(lambda x: not isinstance(x, FilterItem), filters)), \
            'All filters must be instances of FilterItem'

        assert isinstance(sorting, (list, tuple, SortItem)), 'Sorting must be a list or tuple'
        if isinstance(sorting, SortItem):
            sorting = (sorting, )

        assert not list(filter(lambda x: not isinstance(x, SortItem), sorting)), \
            'All sorting rules must be instances of SortItem'

        result = self.__request("SearchTransportOrders", "POST", {
            'context': context,
            'filters': filters,
            'sorting': sorting,
            'start': start,
            'limit': limit,
            'show_count': show_count,
        })

        if result.get("Success", False):
            return TransportOrderListPage(dict_data=result["raw"]["SearchTransportOrdersResult"])
        else:
            raise ElvisException(result['message'], result['raw'])

    def get_transport_order(self, transport_order_id):
        assert self.session_token, "No valid session available"

        result = self.__request("GetTransportOrder", "POST", {
            'transport_order_id': transport_order_id,
        })

        if result.get("Success", False):
            json_obj = result["raw"]["GetTransportOrderResult"]
            return TransportOrder(dict_data=json_obj)
        else:
            raise ElvisException(result['message'], result['raw'])

    def get_transport_order_status(self, transport_order_id):
        assert self.session_token, "No valid session available"

        result = self.__request("GetTransportOrderStatus", "POST", {
            'transport_order_id': transport_order_id
        })

        if result.get("Success", False):
            return TransportOrderStatusInfo(dict_data=result["raw"]["GetTransportOrderStatusResult"])
        else:
            raise ElvisException(result['message'], result['raw'])

    def set_transport_order_status(self, transport_order_id, status, feedback, version):
        assert self.session_token, "No valid session available"

        result = self.__request("SetTransportOrderStatus", "POST", {
            'transport_order_id': transport_order_id,
            'status': status,
            'feedback': feedback or "",
            'version': version or [],
        })

        if result.get("Success", False):
            return result["raw"]['SetTransportOrderStatusResult']
        else:
            raise ElvisException(result['message'], result['raw'])

    def insert_warehouse(self, warehouse):
        assert self.session_token, "No valid session available"

        assert isinstance(warehouse, TimberWarehouse), 'Invalid Warehouse'

        result = self.__request("InsertWarehouse", "POST", {
            'item': warehouse,
        })

        if result.get("Success", False):
            return result["raw"]['InsertWarehouseResult']
        else:
            raise ElvisException(result['message'], result['raw'])

    def get_warehouse(self, warehouse_id):
        assert self.session_token, "No valid session available"

        result = self.__request("GetWarehouse", "POST", {
            'warehouse_id': warehouse_id,
        })

        if result.get("Success", False):
            json_obj = result["raw"]["GetWarehouseResult"]
            return TimberWarehouse(dict_data=json_obj)
        else:
            raise ElvisException(result['message'], result['raw'])

    def delete_warehouse(self, warehouse_id):
        assert self.session_token, "No valid session available"

        result = self.__request("DeleteWarehouse", "POST", {
            'warehouse_id': warehouse_id,
        })

        if result.get("Success", False):
            return result["raw"]['DeleteWarehouseResult']
        else:
            raise ElvisException(result['message'], result['raw'])

    def insert_waybill(self, waybill):
        assert self.session_token, "No valid session available"

        assert isinstance(waybill, Waybill), 'Invalid Waybill'

        result = self.__request("InsertWaybill", "POST", {
            'item': waybill,
        })

        if result.get("Success", False):
            return result["raw"]["InsertWaybillResult"]
        else:
            raise ElvisException(result['message'], result['raw'])

    def get_waybill(self, waybill_id):
        assert self.session_token, "No valid session available"

        result = self.__request("GetWaybill", "POST", {
            'waybill_id': waybill_id
        })

        if result.get("Success", False):
            json_obj = result["raw"]["GetWaybillResult"]
            if json_obj is None:
                return None
            return Waybill(dict_data=json_obj)
        else:
            raise ElvisException(result['message'], result['raw'])

    def set_waybill_status(self, waybill_number, status, feedback, pre_journey_length,
                           total_journey_length, measurement_act_nr, version):
        assert self.session_token, "No valid session available"

        result = self.__request("SetWaybillStatus", "POST", {
            'waybillNumber': waybill_number,
            'status': status,
            'feedback': feedback,
            'preJorneyLength': pre_journey_length,
            'totalJourneyLength': total_journey_length,
            'measurementActNr': measurement_act_nr,
            'version': version,
        })

        if result.get("Success", False):
            return result["raw"]['SetWaybillStatusResult']
        else:
            raise ElvisException(result['message'], result['raw'])

    def get_waybill_status(self, waybill_id):
        assert self.session_token, "No valid session available"

        result = self.__request("GetWaybillStatus", "POST", {
            'waybill_id': waybill_id
        })

        if result.get("Success", False):
            return WaybillStatusInfo(dict_data=result["raw"]["GetWaybillStatusResult"])
        else:
            raise ElvisException(result['message'], result['raw'])

    def search_waybills(self, context, filters, sorting, start=0, limit=10, show_count=False):
        assert self.session_token, "No valid session available"

        assert context in WaybillRoleContext.__dict__.values(), 'Invalid context'

        assert isinstance(filters, (list, tuple, FilterItem)), 'Filters must be a list or tuple'
        if isinstance(filters, FilterItem):
            filters = (filters, )

        assert not list(filter(lambda x: not isinstance(x, FilterItem), filters)), \
            'All filters must be instances of FilterItem'

        assert isinstance(sorting, (list, tuple, SortItem)), 'Sorting must be a list or tuple'
        if isinstance(sorting, SortItem):
            sorting = (sorting, )

        assert not list(filter(lambda x: not isinstance(x, SortItem), sorting)), \
            'All sorting rules must be instances of SortItem'

        assert filters and sorting, 'Need to add filters and sorting!'

        result = self.__request("SearchWaybills", "POST", {
            'context': context,
            'filters': filters,
            'sorting': sorting,
            'start': start,
            'limit': limit,
            'show_count': show_count,
        })

        if result.get("Success", False):
            return WaybillListPage(dict_data=result["raw"]["SearchWaybillsResult"])
        else:
            raise ElvisException(result['message'], result['raw'])

    def insert_reception_assortment(self, timber_batch_id, timber_assortment):
        assert self.session_token, "No valid session available"

        assert isinstance(timber_assortment, TimberAssortment), 'Invalid TimberAssortment'

        result = self.__request("InsertReceptionAssortment", "POST", {
            'timber_batch_id': timber_batch_id,
            'assortment': timber_assortment,
        })

        if result.get("Success", False):
            return result["raw"]["InsertReceptionAssortmentResult"]
        else:
            raise ElvisException(result['message'], result['raw'])

    def delete_reception_assortment(self, reception_assortment_id):
        return self.__delete_function({'id': reception_assortment_id}, 'DeleteReceptionAssortment')

    def insert_fine_measurement_assortment(self, waybill_number, timber_assortment):
        assert self.session_token, "No valid session available"

        assert isinstance(timber_assortment, TimberAssortment), 'Invalid TimberAssortment'

        result = self.__request("InsertFineMeasurementAssortment", "POST", {
            'waybill_number': waybill_number,
            'assortment': timber_assortment,
        })

        if result.get("Success", False):
            val = result["raw"]["InsertFineMeasurementAssortmentResult"]

            if not val or val in ['0', 0]:
                raise ElvisException('InsertFineMeasurementAssortmentResult:: returned id 0', result['raw'])

            else:
                return val
        else:
            raise ElvisException(result['message'], result['raw'])

    def delete_fine_measurement_assortment(self, reception_assortment_id):
        return self.__delete_function({'id': reception_assortment_id}, 'DeleteFineMeasurementAssortment')

    def insert_fine_measurement_file(self, waybill_number, fine_measurement_file):
        assert self.session_token, "No valid session available"

        assert isinstance(fine_measurement_file, FineMeasurementFile), 'Invalid FineMeasurementFile'

        result = self.__request("InsertFineMeasurementFile", "POST", {
            'waybill_number': waybill_number,
            'file': fine_measurement_file,
            'file_data': force_text(base64.b64encode(fine_measurement_file.Data)),
        })

        if result.get("Success", False):
            return result["raw"]["InsertFineMeasurementFileResult"]
        else:
            raise ElvisException(result['message'], result['raw'])

    def delete_fine_measurement_file(self, fine_measurement_file_id):
        return self.__delete_function({'id': fine_measurement_file_id}, 'DeleteFineMeasurementFile')

    def __delete_function(self, params, endpoint):
        assert self.session_token, "No valid session available"

        result = self.__request(endpoint, "POST", params)

        if result.get("Success", False):
            return result["raw"]["%sResult" % endpoint]
        else:
            raise ElvisException(result['message'], result['raw'])



class ElvisEnum(object):
    @classmethod
    def display_name(cls, status):
        for item in cls.__dict__:
            if not item.startswith("_") and isinstance(cls.__dict__[item], int) and cls.__dict__[item] == int(status):
                return item
        return "Unknown"

    @classmethod
    def get_status_choices(cls):
        choices = []
        for item in cls.__dict__:
            if not item.startswith("_") and isinstance(cls.__dict__[item], int):
                choices.append((cls.__dict__[item], item))

        return choices


class AssortmentType(ElvisEnum):
    ELVIS = 0
    COMPANY = 1


class Priority(ElvisEnum):
    Low = 11001
    Normal = 11002
    High = 11003


class WarehouseListItemSearchField(ElvisEnum):
    CompanyRegistrationNumber = 0
    UserId = 1
    Code = 2
    Name = 3
    TypeId = 4
    Address = 5
    ContactName = 6
    CreatedOnStart = 7
    CreatedOnEnd = 8
    IsActive = 9
    IsPublic = 10
    IsArchived = 11
    AssortmentName = 12
    HoldingBaseId = 13
    TimberBatchContractNumber = 14
    TimberBatchDocumentDate = 15
    WaybillNumber = 16
    WaybillNumberAlternate = 17
    TransportOrderNumber = 18
    TransportOrderNumberAlternate = 19
    WaybillStatusId = 20
    AdditionalPropertyTypeId = 21
    AdditionalProperty = 22


class TransportOrderListItemSearchField(ElvisEnum):
    GroupId = 0
    OwnerCode = 1
    RecieverCode = 2
    TransporterCode = 3
    DeadlineFrom = 4
    DeadlineTo = 5
    TransportOrderNumber = 6
    TransportOrderNumberAlternate = 7
    OwnerCompanyName = 8
    TransporterCompanyName = 9
    SourceWarehouseName = 10
    DestinationWarehouseName = 11
    IsDeleted = 12
    StatusId = 13
    IsArchived = 14
    ReportFilterDateFrom = 15
    ReportFilterDateTo = 16
    RecieverName = 17
    SourceWarehouseCodeOrName = 18
    DestinationWarehouseCodeOrName = 19
    AssortmentName = 20
    HoldingBaseId = 21
    TimberBatchContractNumber = 22
    TimberBatchDocumentDate = 23
    DriverName = 24
    VanRegistrationNumber = 25
    DriverPersonCode = 26
    VisibleToReceiver = 27
    AdditionalPropertyTypeId = 28
    AdditionalProperty = 29
    ContactType = 30
    ContactName = 31


class TransportOrderListItemSortField(ElvisEnum):
    Deadline = 0
    CreatedOn = 1
    OwnerName = 2
    DestinationWarehouseName = 3
    TransporterName = 4
    Number = 5


class WarehouseListItemSortField(ElvisEnum):
    Code = 0
    Name = 1
    TypeId = 2
    Address = 3
    ResponsiblePersonName = 4
    CreatedOn = 5


class SortDirection(ElvisEnum):
    Asc = 0
    Desc = 1


class TransportOrderRoleContext(ElvisEnum):
    Owner = 0
    Reciever = 1
    Transporter = 2


class WaybillRoleContext(ElvisEnum):
    Owner = 1
    Reciever = 2
    Transporter = 4
    All = 8


class WarehouseType(ElvisEnum):
    ForestEdge = 5001
    Middle = 5002
    End = 5003


class VehicleType(ElvisEnum):
    Van = 9001
    Trailer = 9002


class WaybillStatus(ElvisEnum):
    Composing = 7001
    Confirmed = 7002
    Unloaded = 7003
    Received = 7004
    Measured = 7005
    MeasurementConfirmed = 7006
    Finalized = 7007
    Cancelled = 7008


class WaybillListItemSearchField(ElvisEnum):
    OwnerCode = 0
    RecieverCode = 1
    RecieverName = 2
    TransporterCode = 3
    WaybillNumber = 4
    WaybillNumberAlternate = 5
    OwnerCompanyName = 6
    TransporterCompanyName = 7
    SourceWarehouseName = 8
    SourceWarehouseCodeOrName = 9
    DestinationWarehouseName = 10
    DestinationWarehouseCodeOrName = 11
    IsDeleted = 12
    StatusId = 13
    IsArchived = 14
    ReportFilterDateFrom = 15
    ReportFilterDateTo = 16
    CreatedOnStart = 17
    CreatedOnEnd = 18
    TransportOrderNumber = 19
    TransportOrderNumberAlternate = 20
    AssortmentName = 21
    HoldingBaseId = 22
    TimberBatchContractNumber = 23
    TimberBatchDocumentDate = 24
    DriverName = 25
    VanRegistrationNumber = 26
    TrailerRegistrationNumber = 27
    IsDisputed = 28
    DriverPersonCode = 29
    ReportCompanyName = 30
    AdditionalPropertyTypeId = 31
    AdditionalProperty = 32
    ContactType = 33
    ContactName = 34
    AuthorizedReceiverCode = 35
    NotInStatusId = 36
    Inspected = 37


class WaybillListItemSortField(ElvisEnum):
    CreatedOn = 0
    OwnerName = 1
    DestinationWarehouseName = 2
    TransporterName = 3
    Number = 4


class TransportOrderStatus(ElvisEnum):
    Composing = 4001
    Submitted = 4002
    Accepted = 4003
    Declined = 4004
    Finalized = 4005
    Cancelled = 4006



import hashlib
import json

from django.utils.encoding import force_bytes

from .enums import Priority, TransportOrderStatus
from .utils import decode_elvis_timestamp


class FilterItem(object):
    def __init__(self, field, value):
        self.Field = field
        self.Value = value


class SortItem(object):
    def __init__(self, column, direction):
        self.SortColumn = column
        self.SortDirection = direction


class ElvisModel(object):
    _DATETIME_ATTRIBUTES = []

    def __init__(self, **kwargs):
        dict_data = kwargs.get('dict_data', None)
        if dict_data:
            self.__dict__ = dict_data.copy()
            self._already_loaded = True
        else:
            self._already_loaded = False

    def __getattribute__(self, name):
        if name.startswith("_"):
            return super().__getattribute__(name)

        if name not in self._DATETIME_ATTRIBUTES:
            return super().__getattribute__(name)

        return decode_elvis_timestamp(super().__getattribute__(name))


# noinspection PyPep8Naming
class AdditionalProperty(ElvisModel):
    def __init__(self, **kwargs):
        super(AdditionalProperty, self).__init__(**kwargs)
        if not self._already_loaded:
            self.TypeId = kwargs.get('type_id')
            self.Value = kwargs.get('value')

            self.ExtensionData = None


# noinspection PyPep8Naming
class Address(ElvisModel):
    def __init__(self, **kwargs):
        super(Address, self).__init__(**kwargs)
        if not self._already_loaded:
            self.AddressAdditionalInformation = kwargs.get('address_additional_information')
            self.CityBorough = kwargs.get('city_borough')
            self.County = kwargs.get('county')
            self.EHAK = kwargs.get('ehak')
            self.NearAddress = kwargs.get('near_address')
            self.RuralDistrict = kwargs.get('rural_district')


# noinspection PyPep8Naming
class Pack(ElvisModel):
    def __init__(self, **kwargs):
        super(Pack, self).__init__(**kwargs)

        if not self._already_loaded:
            self.Factor = kwargs.get('factor')
            self.Width = kwargs.get('width')
            self.Heidht = kwargs.get('height')
            self.Length = kwargs.get('length')
            self.Number = kwargs.get('number')
            self.VehicleType = kwargs.get('vehicle_type')

            self.ExtensionData = None


# noinspection PyPep8Naming
class TimberAssortment(ElvisModel):
    def __init__(self, **kwargs):
        super(TimberAssortment, self).__init__(**kwargs)

        if not self._already_loaded:
            self.Amount = kwargs.get('amount')
            self.Description = kwargs.get('description')
            self.TimberAssortmentTypeId = kwargs.get('assortment_type')

            self.Id = kwargs.get('timber_assortment_id', None)

            # Load Packs
            packs = kwargs.get('packs', [])
            if isinstance(packs, Pack):
                packs = [packs, ]

            assert not list(filter(lambda x: not isinstance(x, Pack), packs))
            self.Packs = packs

            self.ExtensionData = None
        else:
            # Load received assortments
            packs = self.Packs
            self.Packs = []

            if packs:
                for pack in packs:
                    self.Packs.append(Pack(dict_data=pack))

        assert self.Amount is not None, 'TimberAssortment::Amount cant be None'


# noinspection PyPep8Naming
class Certificate(ElvisModel):
    def __init__(self, **kwargs):
        super(Certificate, self).__init__(**kwargs)

        if not self._already_loaded:
            self.CertificateNumber = kwargs.get('certificate_number')
            self.TypeId = kwargs.get('type_id')
            self.ExtensionData = None


# noinspection PyPep8Naming
class TimberBatch(ElvisModel):
    _DATETIME_ATTRIBUTES = ["DocDate"]

    def __init__(self, **kwargs):
        super(TimberBatch, self).__init__(**kwargs)

        if not self._already_loaded:
            self.Appropriation = kwargs.get('appropriation')  # string
            self.Assortments = kwargs.get('assortments')  # TimberAssortment[]
            self.Certificates = kwargs.get('certificates')  # Certificate[]
            self.Description = kwargs.get('description')  # string

            self.CadastralNumber = kwargs.get('cadastral_number')  # string

            self.DocDate = kwargs.get('doc_date')  # DateTime
            self.DocNumber = kwargs.get('doc_number')  # string

            self.HoldingBaseId = kwargs.get('holding_base_id')  # int
            self.Id = kwargs.get('batch_id', None)  # int or None
            self.ForestNotice = kwargs.get('forest_notice', None)  # string or None

            self.PreviousOwnerAddress = kwargs.get('prev_owner_address')  # string
            self.PreviousOwnerCode = kwargs.get('prev_owner_code')  # string
            self.PreviousOwnerName = kwargs.get('prev_owner_name')  # string

            self.Quarter = kwargs.get('quarter')  # string
            self.RegisteredImmovableNumber = kwargs.get('reg_immovable_number')  # string

            self.ExtensionData = None
        else:
            assortments = self.Assortments
            certs = self.Certificates
            self.Assortments = []
            self.Certificates = []

            for item in assortments:
                self.Assortments.append(TimberAssortment(dict_data=item))
            for item in certs:
                self.Certificates.append(Certificate(dict_data=item))


# noinspection PyPep8Naming
class Warehouse(ElvisModel):
    def __init__(self, **kwargs):
        super(Warehouse, self).__init__(**kwargs)

        if not self._already_loaded:
            self.Id = kwargs.get('warehouse_id', None)  # string
            self.Name = kwargs.get('name')  # string
            self.Code = kwargs.get('code')  # string

            self.ContactEmail = kwargs.get('contact_email')  # string
            self.ContactName = kwargs.get('contact_name')  # string
            self.ContactPhone = kwargs.get('contact_phone')  # string

            self.Description = kwargs.get('description')  # string
            self.Appropriation = kwargs.get('appropriation')  # string
            self.Quarter = kwargs.get('quarter')  # string

            self.IsDry = kwargs.get('is_dry')  # bool or None
            self.IsParallelLoading = kwargs.get('is_parallel_loading')  # bool or None

            self.LambertEstX = kwargs.get('lambert_est_x')  # double or None
            self.LambertEstY = kwargs.get('lambert_est_y')  # double or None

            assert isinstance(kwargs.get('address'), Address)
            self.Address = kwargs.get('address')

            self.ForestDistrictId = kwargs.get('forest_district_id')  # int or None

            additional_properties = kwargs.get('additional_properties', [])
            if isinstance(additional_properties, AdditionalProperty):
                additional_properties = [additional_properties, ]

            assert not list(filter(lambda x: not isinstance(x, AdditionalProperty), additional_properties))
            self.AdditionalProperties = additional_properties
        else:
            if not isinstance(self.Address, Address):
                self.Address = Address(dict_data=self.Address)

            props = self.AdditionalProperties
            self.AdditionalProperties = []

            if props:
                for item in props:
                    self.AdditionalProperties.append(AdditionalProperty(dict_data=item))


# noinspection PyPep8Naming
class TimberWarehouse(Warehouse):

    def __init__(self, **kwargs):

        super(TimberWarehouse, self).__init__(**kwargs)

        #if not self._ElvisModel_already_loaded:
        if not self._already_loaded:
            self.CadastralNumber = kwargs.get('cadastral_number')
            self.IsActive = kwargs.get('is_active')
            self.IsPublic = kwargs.get('is_public')

            timber_batches = kwargs.get('timber_batches')
            if isinstance(timber_batches, TimberBatch):
                timber_batches = [timber_batches, ]

            assert not list(filter(lambda x: not isinstance(x, TimberBatch), timber_batches))
            self.TimberBatches = timber_batches

            self.Type = kwargs.get('warehouse_type')  # WarehouseType

            self.ExtensionData = None
            self.Version = kwargs.get('version', None)  # Note: byte array
        else:
            batches = self.TimberBatches
            self.TimberBatches = []

            for batch in batches:
                self.TimberBatches.append(TimberBatch(dict_data=batch))


# noinspection PyPep8Naming
class WaybillStatusChangeLog(ElvisModel):
    _DATETIME_ATTRIBUTES = ["ChangedOn"]

    def __init__(self, **kwargs):

        super(WaybillStatusChangeLog, self).__init__(**kwargs)

        if not self._already_loaded:
            self.ChangedBy = kwargs.get('changed_by')
            self._ChangedOn = kwargs.get('changed_on')
            self.Status = kwargs.get('status')

            self.ExtensionData = None


# noinspection PyPep8Naming
class Shipment(ElvisModel):
    def __init__(self, **kwargs):

        super(Shipment, self).__init__(**kwargs)

        if not self._already_loaded:
            assert isinstance(kwargs.get('warehouse'), Warehouse)
            self.Warehouse = kwargs.get('warehouse')

            timber_batches = kwargs.get('timber_batches')
            if isinstance(timber_batches, TimberBatch):
                timber_batches = [timber_batches, ]

            assert not list(filter(lambda x: not isinstance(x, TimberBatch), timber_batches))
            self.TimberBatches = timber_batches

            self.ExtensionData = None
        else:
            if not isinstance(self.Warehouse, Warehouse):
                self.Warehouse = Warehouse(dict_data=self.Warehouse)

            batches = self.TimberBatches
            self.TimberBatches = []

            for batch in batches:
                self.TimberBatches.append(TimberBatch(dict_data=batch))


# noinspection PyPep8Naming
class Person(ElvisModel):
    def __init__(self, **kwargs):
        super(Person, self).__init__(**kwargs)

        if not self._already_loaded:
            self.EMail = kwargs.get('email')
            self.Firstname = kwargs.get('firstname')
            self.Lastname = kwargs.get('lastname')
            self.PersonCode = kwargs.get('person_code')
            self.Phone = kwargs.get('phone')

            self.ExtensionData = None


# noinspection PyPep8Naming
class AuthorizedPerson(Person):
    def __init__(self, **kwargs):
        super(AuthorizedPerson, self).__init__(**kwargs)

        if not self._already_loaded:
            assert isinstance(kwargs.get('address'), Address)
            self.Address = kwargs.get('address')
            self.AuthorizationBase = kwargs.get('authorization_base')
        else:
            if not isinstance(self.Address, Address):
                self.Address = Address(dict_data=self.Address)


# noinspection PyPep8Naming
class Party(ElvisModel):
    def __init__(self, **kwargs):
        super(Party, self).__init__(**kwargs)

        if not self._already_loaded:
            self.Code = kwargs.get('code')
            self.Email = kwargs.get('email')
            self.Name = kwargs.get('name')
            self.Phone = kwargs.get('phone')

            self.ExtensionData = None


# noinspection PyPep8Naming
class TimberOwner(Party):
    def __init__(self, **kwargs):
        super(TimberOwner, self).__init__(**kwargs)

        if not self._already_loaded:
            self.Address = kwargs.get('address')

            self.AuthorizedPerson = kwargs.get('authorized_person')
            if self.AuthorizedPerson is not None:
                assert isinstance(self.AuthorizedPerson, AuthorizedPerson)

        else:
            if not isinstance(self.AuthorizedPerson, AuthorizedPerson):
                self.AuthorizedPerson = AuthorizedPerson(dict_data=self.AuthorizedPerson)


# noinspection PyPep8Naming
class TimberReceiver(Party):
    def __init__(self, **kwargs):
        super(TimberReceiver, self).__init__(**kwargs)

        if not self._already_loaded:
            self.ContactEmail = kwargs.get('contact_email')
            self.ContactName = kwargs.get('contact_name')
            self.ContactPhone = kwargs.get('contact_phone')

            self.Address = kwargs.get('address')
            if self.Address is not None:
                assert isinstance(self.Address, Address)
            self.AuthorizationBase = kwargs.get('authorization_base')

            additional_properties = kwargs.get('additional_properties', [])
            if isinstance(additional_properties, AdditionalProperty):
                additional_properties = [additional_properties, ]

            assert not list(filter(lambda x: not isinstance(x, AdditionalProperty), additional_properties))
            self.AdditionalProperties = additional_properties
        else:
            if not isinstance(self.Address, Address):
                self.Address = Address(dict_data=self.Address)

            props = self.AdditionalProperties
            self.AdditionalProperties = []

            if props:
                for item in props:
                    self.AdditionalProperties.append(AdditionalProperty(dict_data=item))


# noinspection PyPep8Naming
class TimberReceiverDestination(ElvisModel):
    def __init__(self, **kwargs):
        super(TimberReceiverDestination, self).__init__(**kwargs)

        if not self._already_loaded:
            assert isinstance(kwargs.get('destination'), Warehouse)
            self.Destination = kwargs.get('destination')

            assert isinstance(kwargs.get('receiver'), TimberReceiver)
            self.Receiver = kwargs.get('receiver')

            self.ExtensionData = None
        else:
            if not isinstance(self.Destination, Warehouse):
                self.Destination = Warehouse(dict_data=self.Destination)

            if not isinstance(self.Receiver, TimberReceiver):
                self.Receiver = TimberReceiver(dict_data=self.Receiver)


# noinspection PyPep8Naming
class Vehicle(ElvisModel):
    def __init__(self, **kwargs):
        super(Vehicle, self).__init__(**kwargs)

        if not self._already_loaded:
            self.Model = kwargs.get('model')
            self.RegistrationNumber = kwargs.get('registration_number')

            self.ExtensionData = None


# noinspection PyPep8Naming
class Transport(ElvisModel):
    def __init__(self, **kwargs):
        super(Transport, self).__init__(**kwargs)

        if not self._already_loaded:
            assert isinstance(kwargs.get('driver'), Person)
            self.Driver = kwargs.get('driver')

            assert isinstance(kwargs.get('trailer'), Vehicle)
            self.Trailer = kwargs.get('trailer')

            assert isinstance(kwargs.get('van'), Vehicle)
            self.Van = kwargs.get('van')

            self.ExtensionData = None
        else:
            if not isinstance(self.Driver, Person):
                self.Driver = Person(dict_data=self.Driver)

            if not isinstance(self.Trailer, Vehicle):
                self.Trailer = Vehicle(dict_data=self.Trailer)

            if not isinstance(self.Van, Vehicle):
                self.Van = Vehicle(dict_data=self.Van)


# noinspection PyPep8Naming
class Transporter(ElvisModel):
    def __init__(self, **kwargs):
        super(Transporter, self).__init__(**kwargs)

        if not self._already_loaded:
            self.CompanyRegistrationNumber = kwargs.get('company_registration_number')
            self.ContactEmail = kwargs.get('contact_email')
            self.ContactName = kwargs.get('contact_name')
            self.ContactPhone = kwargs.get('contact_phone')

            self.ExtensionData = None

            additional_properties = kwargs.get('additional_properties', [])
            if isinstance(additional_properties, AdditionalProperty):
                additional_properties = [additional_properties, ]

            assert not list(filter(lambda x: not isinstance(x, AdditionalProperty), additional_properties))
            self.AdditionalProperties = additional_properties
        else:
            props = self.AdditionalProperties
            self.AdditionalProperties = []

            if props:
                for item in props:
                    self.AdditionalProperties.append(AdditionalProperty(dict_data=item))


# noinspection PyPep8Naming
class WaybillTransporter(Transporter):

    def __init__(self, **kwargs):
        super(WaybillTransporter, self).__init__(**kwargs)

        if not self._already_loaded:
            assert isinstance(kwargs.get('transport'), Transport)
            self.Transport = kwargs.get('transport')
        else:
            if not isinstance(self.Transport, Transport):
                self.Transport = Transport(dict_data=self.Transport)


# noinspection PyPep8Naming
class Waybill(ElvisModel):
    def __init__(self, **kwargs):
        super(Waybill, self).__init__(**kwargs)

        if not self._already_loaded:
            self.AltNumber = kwargs.get('alt_number')
            self.Description = kwargs.get('description')
            self.IsDisputed = kwargs.get('is_disputed', False)
            self.Number = kwargs.get('number')
            self.Status = kwargs.get('status')

            self.TransportOrderNumber = kwargs.get('transport_order_number')

            self.PreJourneyLength = kwargs.get('pre_journey_length', None)
            self.TotalJourneyLength = kwargs.get('total_journey_length', None)

            self.Version = kwargs.get('version', None)  # Note: byte array
            self.ExtensionData = None

            # Timber owner
            if not isinstance(kwargs.get('timber_owner'), TimberOwner):
                from elvis.api import ElvisException
                raise ElvisException("ELVIS didn't return a proper way bill", raw={})
            self.TimberOwner = kwargs.get('timber_owner')

            # Timber receiver destination
            assert isinstance(kwargs.get('timber_receiver_destination'), TimberReceiverDestination)
            self.TimberReceiverDestination = kwargs.get('timber_receiver_destination')

            # Transporter
            assert isinstance(kwargs.get('transporter'), WaybillTransporter)
            self.Transporter = kwargs.get('transporter')

            # Load fine measurements
            fine_measurements = kwargs.get('fine_measurements')
            if fine_measurements:
                if isinstance(fine_measurements, TimberAssortment):
                    fine_measurements = [fine_measurements, ]

                assert not list(filter(lambda x: not isinstance(x, TimberAssortment), fine_measurements))
                self.FineMeasurements = fine_measurements

            # Load received assortments
            received_assortments = kwargs.get('received_assortments')
            if received_assortments:
                if isinstance(received_assortments, TimberBatch):
                    received_assortments = [received_assortments, ]

                assert not list(filter(lambda x: not isinstance(x, TimberBatch), received_assortments))
                self.ReceivedAssortments = received_assortments

            # Load status change logs
            status_change_logs = kwargs.get('status_change_logs')
            if status_change_logs:
                if isinstance(status_change_logs, WaybillStatusChangeLog):
                    status_change_logs = [status_change_logs, ]

                assert not list(filter(lambda x: not isinstance(x, WaybillStatusChangeLog), status_change_logs))
                self.StatusChangeLogs = status_change_logs

            # Load shipments
            shipments = kwargs.get('shipments')
            if shipments:
                if isinstance(shipments, Shipment):
                    shipments = [shipments, ]

                assert not list(filter(lambda x: not isinstance(x, Shipment), shipments))
                self.Shipments = shipments
        else:
            # Timber owner
            if not isinstance(self.TimberOwner, TimberOwner):
                self.TimberOwner = TimberOwner(dict_data=self.TimberOwner)

            # Timber receiver destination
            if not isinstance(self.TimberReceiverDestination, TimberReceiverDestination):
                self.TimberReceiverDestination = TimberReceiverDestination(dict_data=self.TimberReceiverDestination)

            # Transporter
            if not isinstance(self.Transporter, WaybillTransporter):
                self.Transporter = WaybillTransporter(dict_data=self.Transporter)

            # Load fine measurements
            fine_measurements = self.FineMeasurements
            self.FineMeasurements = []

            if fine_measurements:
                for fine_measurement in fine_measurements:
                    self.FineMeasurements.append(TimberAssortment(dict_data=fine_measurement))

            # Load received assortments
            received_assortments = self.ReceivedAssortments
            self.ReceivedAssortments = []

            if received_assortments:
                for received_assortment in received_assortments:
                    self.ReceivedAssortments.append(TimberBatch(dict_data=received_assortment))

            # Load status change logs
            status_change_logs = self.StatusChangeLogs
            self.StatusChangeLogs = []

            if status_change_logs:
                for item in status_change_logs:
                    self.StatusChangeLogs.append(WaybillStatusChangeLog(dict_data=item))

            # Load shipments
            shipments = self.Shipments
            self.Shipments = []

            if shipments:
                for item in shipments:
                    self.Shipments.append(Shipment(dict_data=item))


# noinspection PyPep8Naming
class WaybillListItem(ElvisModel):
    def __init__(self, **kwargs):
        super(WaybillListItem, self).__init__(**kwargs)

        if not self._already_loaded:
            self.Number = kwargs.get('number')
            self.AltNumber = kwargs.get('alt_number')

            self.OwnerCode = kwargs.get('owner_code')
            self.OwnerName = kwargs.get('owner_name')

            self.ReceiverCode = kwargs.get('receiver_code')
            self.ReceiverName = kwargs.get('receiver_name')

            self.Status = kwargs.get('status')

            self.TransportOrderNumber = kwargs.get('transport_order_number')

            self.TransporterCode = kwargs.get('transporter_code')
            self.TransporterName = kwargs.get('transporter_name')

            self.Version = kwargs.get('version', [])

            self.ExtensionData = None


# noinspection PyPep8Naming
class TransportOrderListItem(ElvisModel):
    def __init__(self, **kwargs):
        super(TransportOrderListItem, self).__init__(**kwargs)

        if not self._already_loaded:
            self.Number = kwargs.get('number')
            self.AltNumber = kwargs.get('alt_number')

            self.OwnerCode = kwargs.get('owner_code')
            self.OwnerName = kwargs.get('owner_name')

            self.ReceiverCode = kwargs.get('receiver_code')
            self.ReceiverName = kwargs.get('receiver_name')

            self.TransporterCode = kwargs.get('transporter_code')
            self.TransporterName = kwargs.get('transporter_name')

            self.Deadline = kwargs.get('deadline', None)
            self.GroupId = kwargs.get('group_id')

            self.Version = kwargs.get('version', [])
            self.ExtensionData = None

            self.Status = kwargs.get('status')
            self.Priority = kwargs.get('priority')


# noinspection PyPep8Naming
class WaybillListPage(ElvisModel):
    def __init__(self, **kwargs):
        super(WaybillListPage, self).__init__(**kwargs)

        if not self._already_loaded:
            self.HasCount = kwargs.get('has_count')
            self.Skip = kwargs.get('skip')
            self.Take = kwargs.get('take')
            self.TotalCount = kwargs.get('total_count')

            self.ExtensionData = None

            # Load shipments
            items = kwargs.get('items', [])
            if isinstance(items, WaybillListItem):
                items = [items, ]

            assert not list(filter(lambda x: not isinstance(x, WaybillListItem), items))
            self.Items = items

        else:
            # Load items
            items = self.Items
            self.Items = []

            if items:
                for item in items:
                    self.Items.append(WaybillListItem(dict_data=item))


# noinspection PyPep8Naming
class TransportOrderListPage(ElvisModel):
    def __init__(self, **kwargs):
        super(TransportOrderListPage, self).__init__(**kwargs)

        if not self._already_loaded:
            self.HasCount = kwargs.get('has_count')
            self.Skip = kwargs.get('skip')
            self.Take = kwargs.get('take')
            self.TotalCount = kwargs.get('total_count')

            self.ExtensionData = None

            # Load shipments
            items = kwargs.get('items', [])
            if isinstance(items, TransportOrderListItem):
                items = [items, ]

            assert not list(filter(lambda x: not isinstance(x, TransportOrderListItem), items))
            self.Items = items

        else:
            # Load items
            items = self.Items
            self.Items = []

            if items:
                for item in items:
                    self.Items.append(TransportOrderListItem(dict_data=item))


# noinspection PyPep8Naming
class WaybillStatusInfo(ElvisModel):
    def __init__(self, **kwargs):
        super(WaybillStatusInfo, self).__init__(**kwargs)

        if not self._already_loaded:
            self.Number = kwargs.get('number')
            self.Status = kwargs.get('status')
            self.Version = kwargs.get('status', [])
            self.ExtensionData = None


# noinspection PyPep8Naming
class TransportOrderStatusInfo(ElvisModel):
    def __init__(self, **kwargs):
        super(TransportOrderStatusInfo, self).__init__(**kwargs)

        if not self._already_loaded:
            self.Number = kwargs.get('number')
            self.Status = kwargs.get('status')
            self.Version = kwargs.get('status', [])
            self.ExtensionData = None


# noinspection PyPep8Naming
class TransportOrderStatusChangeLog(WaybillStatusChangeLog):
    pass


# noinspection PyPep8Naming
class TransportOrderTransporter(Transporter):
    def __init__(self, **kwargs):
        super(TransportOrderTransporter, self).__init__(**kwargs)

        if not self._already_loaded:

            transports = kwargs.get('transports', [])
            if isinstance(transports, Transport):
                transports = [transports, ]

            assert transports
            assert not list(filter(lambda x: not isinstance(x, Transport), transports))
            self.Transports = transports

        else:
            transports = self.Transports
            self.Transports = []
            for transport in transports:
                if isinstance(transport, Transport):
                    self.Transports.append(transport)
                else:
                    self.Transports.append(Transport(dict_data=transport))

    def get_waybill_transporter(self, transport_hash):
        transport = None
        for t in self.Transports:
            if transport_hash == hashlib.md5(
                ("%s%s%s" % (t.Driver.PersonCode, t.Trailer.RegistrationNumber, t.Van.RegistrationNumber)).encode("utf-8")
            ).hexdigest():
                transport = t
                break
        assert transport

        from elvis.api import ElvisEncoder
        dict_data = json.loads(json.dumps(self, cls=ElvisEncoder))
        del dict_data['Transports']
        dict_data['Transport'] = transport
        waybill_transport_order = WaybillTransporter(dict_data=dict_data)

        return waybill_transport_order


# noinspection PyPep8Naming
class TransportOrder(ElvisModel):
    _DATETIME_ATTRIBUTES = ["Deadline"]

    def __init__(self, **kwargs):
        super(TransportOrder, self).__init__(**kwargs)

        if not self._already_loaded:
            self.Number = kwargs.get('number')
            self.AltNumber = kwargs.get('alt_number')

            self.Deadline = kwargs.get('deadline', None)
            self.GroupId = kwargs.get('group_id')

            self.Status = kwargs.get('status', TransportOrderStatus.Composing)
            self.Priority = kwargs.get('priority', Priority.Normal)

            self.Description = kwargs.get('description')
            self.TransportContractNumber = kwargs.get('transport_contract_number')

            self.IsPartialOperationsAllowed = kwargs.get('is_partial_operations_allowed')
            self.IsVisibleToReceiver = kwargs.get('is_visible_to_receiver')

            self.Version = kwargs.get('version', [])

            self.ExtensionData = None

            # Timber owner
            if not isinstance(kwargs.get('timber_owner'), TimberOwner):
                from elvis.api import ElvisException
                raise ElvisException("ELVIS didn't return a proper transport order", raw={})
            self.TimberOwner = kwargs.get('timber_owner')

            # Timber receiver destination
            assert isinstance(kwargs.get('timber_receiver_destination'), TimberReceiverDestination)
            self.TimberReceiverDestination = kwargs.get('timber_receiver_destination')

            # Transporter
            assert isinstance(kwargs.get('transporter'), TransportOrderTransporter)
            self.Transporter = kwargs.get('transporter')

            # Load status change logs
            status_change_logs = kwargs.get('status_change_logs')
            if status_change_logs:
                if isinstance(status_change_logs, TransportOrderStatusChangeLog):
                    status_change_logs = [status_change_logs, ]

                assert not list(filter(lambda x: not isinstance(x, TransportOrderStatusChangeLog), status_change_logs))
                self.StatusChangeLogs = status_change_logs

            # Load shipments
            shipments = kwargs.get('shipments')
            if shipments:
                if isinstance(shipments, Shipment):
                    shipments = [shipments, ]

                assert not list(filter(lambda x: not isinstance(x, Shipment), shipments))
                self.Shipments = shipments
        else:
            # Timber owner
            if not isinstance(self.TimberOwner, TimberOwner):
                self.TimberOwner = TimberOwner(dict_data=self.TimberOwner)

            # Timber receiver destination
            if not isinstance(self.TimberReceiverDestination, TimberReceiverDestination):
                self.TimberReceiverDestination = TimberReceiverDestination(dict_data=self.TimberReceiverDestination)

            # Transporter
            if not isinstance(self.Transporter, TransportOrderTransporter):
                self.Transporter = TransportOrderTransporter(dict_data=self.Transporter)

            # Load status change logs
            status_change_logs = self.StatusChangeLogs
            self.StatusChangeLogs = []

            if status_change_logs:
                for item in status_change_logs:
                    self.StatusChangeLogs.append(TransportOrderStatusChangeLog(dict_data=item))

            # Load shipments
            shipments = self.Shipments
            self.Shipments = []

            if shipments:
                for item in shipments:
                    self.Shipments.append(Shipment(dict_data=item))


# noinspection PyPep8Naming
class FineMeasurementFile(ElvisModel):

    def __init__(self, **kwargs):
        super(FineMeasurementFile, self).__init__(**kwargs)

        if not self._already_loaded:
            self.Description = kwargs.get('description')
            self.ContentType = kwargs.get('content_type')
            self.FileName = kwargs.get('file_name')

            self.Data = force_bytes(kwargs.get('data'))

            self.ExtensionData = None

        else:
            self.Data = force_bytes(self.Data)



import os
import sys

try:
    from .local import *
except ImportError:
    PROXY_HOST_URL = os.environ.get('PROXY_HOST_URL')
    TESTING_RANGE = os.environ.get('TESTING_RANGE', True)
    TEST_CERTIFICATE_PATH = os.environ.get('TEST_CERTIFICATE_PATH')
    TEST_CERTIFICATE_PERSON_CODE = os.environ.get('TEST_CERTIFICATE_PERSON_CODE')
    TEST_CERTIFICATE_PASSWORD = os.environ.get('TEST_CERTIFICATE_PASSWORD')
    if None in (PROXY_HOST_URL, TEST_CERTIFICATE_PATH, TEST_CERTIFICATE_PERSON_CODE, TEST_CERTIFICATE_PASSWORD):
        sys.stderr.write("Couldn't import lvis_test_config.local or ENV var not set, have you created it from lvis_test_config/local.py.example?\n")
        sys.exit(1)
